AVERAGE_DURATION = "AverageDuration"
RHYTHMINT = "RhythmInt"  # number of notes / number of beats
# RHYTHMINTSEP = "RhythmIntSep"  # SUM number of notes in a time signature/ beat in a time signature
DOTTEDRHYTHM = "DottedRhythm"
DOUBLE_DOTTEDRHYTHM = "DoubleDottedRhythm"

