ACCIDENTAL_ABBREVIATION = {
    "": "",
    "sharp": "#",
    "flat": "b",
    "double-sharp": "x",
    "double-flat": "bb",
    "triple-flat": "bbb",
    "triple-sharp": "#x"
}
DEGREE_PREFIX = "Degree"
DEGREE_COUNT = "Degree{key}_Count"
DEGREE_PER = "Degree{key}_Per"
