"""Sets the tyoe of encoding to be used"""
ENCODING = "utf-8"

"""Sets the tyoe of encoding to be used"""
CSV_DELIMITER = ","

"""Prefix for voices"""
VOICE_FAMILY = "voice"

GENERAL_FAMILY = "general"
