PART_NAME = "PartName"
PART_NUMBER = "PartNumber"
PART_ABBREVIATION = "PartAbbreviation"
SOUND = "Sound"
SOUND_ABBREVIATION = "SoundAbbreviation"
ABBREVIATION = "abbreviation"
FAMILY = "Family"
FAMILY_ABBREVIATION = "FamilyAbbreviation"
INSTRUMENTAL = "Instrumental"
NUMBER_OF_PARTS = "NumberOfParts"
NUMBER_OF_FILTERED_PARTS = "NumberOfFilteredParts"
NUMBER_OF_PARTS_MEAN = "NumberOfPartsMean"
NUMBER_OF_PARTS_STD = "NumberOfPartsStd"
SCORING = "Scoring"
SOUND_SCORING = "SoundScoring"
INSTRUMENTATION = "Instrumentation"
VOICES = "Voices"
FAMILY_SCORING = "FamilyScoring"
FAMILY_INSTRUMENTATION = "FamilyInstrumentation"
ROLE_TYPE = "RoleType"
