LOWEST_NOTE = "LowestNote"
HIGHEST_NOTE = "HighestNote"
LOWEST_NOTE_INDEX = "LowestNoteIndex"
HIGHEST_NOTE_INDEX = "HighestNoteIndex"
AMBITUS = "Ambitus"
SCORE_FEATURES = [
    LOWEST_NOTE_INDEX,
    LOWEST_NOTE,
    HIGHEST_NOTE_INDEX,
    HIGHEST_NOTE,
    AMBITUS,
]
