DATA_NOTES = "notes"
DATA_NOTES_AND_RESTS = "notes_and_rests"
DATA_LYRICS = "lyrics"
DATA_SOUNDING_MEASURES = "sounding_measures"
DATA_MEASURES = "measures"
DATA_INTERVALS = "intervals"
DATA_KEY = "key"
DATA_KEY_NAME = "key_name"
DATA_MODE = "mode"
FILE_NAME = "FileName"
NUM_NOTES = "Notes"
NOTES_MEAN = "NotesMean"
NUM_SOUNDING_MEASURES = "SoundingMeasures"
SOUNDING_MEASURES_MEAN = "SoundingMeasuresMean"
NUM_MEASURES = "Measures"
KEY_SIGNATURE = "KeySignature"
