DYNMEAN = "DynMean"
DYNMEAN_WEIGHTED = "DynMean_weighted"
DYNABRUPTNESS = "DynAbruptness"
DYNGRAD = "DynGrad"

DYNAMIC = "Dynamic"  # all coulf be in a dictionary with this variables?
TEXTEXPRESSION = "TextExpression"
TIMESIGNATURE = "TimeSignature"
REST = "Rest"

DYNAMIC_VALUES = {
    "sfz": 127,
    "fz": 120,
    "ffff": 116,
    "fff": 112,
    "ff": 101,
    "più f": 96,
    "f assai": 94,
    "assaif": 94,
    "f": 88,
    "sempre f": 88,
    "poco f": 80,
    "mf": 75,
    "mp": 62,
    "p": 49,
    "più p": 46,
    "sempre p": 49,
    "p dolce": 49,
    "poco p": 49,
    "p assai": 42,
    "pp": 36,
    "ppp": 24,
    "sfp": 20,
    "pppp": 12,
    "sempre pp": 36,
    "sotto voce assai": 36,
    "dolce": 49,
}

DYNAMIC_LAST_WORD = ["assai", "dolce"]

DYNAMIC_FIRST_WORD = ["più", "sempre", "poco"]
