ALPHA = "abcdefghijklmnopqrstuvwxyz"
HARMONY_PREFIX = "Harmony_"
HARMONIC_prefix = HARMONY_PREFIX + "Harmonic"
HARMONY_AVAILABLE = HARMONY_PREFIX + "Analysis_Available"

HARMONIC_RHYTHM = HARMONIC_prefix + "Rhythm"
HARMONIC_RHYTHM_BEATS = HARMONIC_prefix + "RhythmBeats"

ADDITIONS_prefix = HARMONY_PREFIX + "Additions_"
ADDITIONS_4_6_64 = "4_6_64"
ADDITIONS_9 = "+9"
OTHERS_NO_AUG = "Others_no_+"
OTHERS_AUG = "Others_+"

KEY_PREFIX = HARMONY_PREFIX + "Key_"
KEY_GROUPING = "KeyGrouping"
KEY_MEASURES = "Measures_"
KEY_PERCENTAGE = "_PercentageMeasures"
KEY_MODULATORY = "Modulatory_"
KEY_MODCOMP = "ModComp_"

NUMERALS_prefix = HARMONY_PREFIX + "Numerals_"

CHORD_TYPES_prefix = HARMONY_PREFIX + "Chord_types_"
CHORD_prefix = HARMONY_PREFIX + "Chord_"
CHORDS_GROUPING_prefix = HARMONY_PREFIX + "Chords_Grouping_"
